@@ -13,5 +13,8 @@ fn main() -> std::io::Result<()> {
         .expect("version field is present in pyproject.toml");
     println!("cargo::rustc-env=PACKAGE_VERSION={}", version);
 
+    // Figure out which Python version we're building for
+    pyo3_build_config::use_pyo3_cfgs();
+
     Ok(())
 }
