@@ -1,6 +1,5 @@
 import argparse
 import os
-import shutil
 import sys
 import tempfile
 from importlib import machinery, util
@@ -20,19 +19,6 @@ exec $(which {python}) -m functiontrace "$@"
 
 
 def setup_dependencies():
-    # We need the functiontrace-server installed and locatable in order to
-    # trace anything.
-    if shutil.which("functiontrace-server") is None:
-        print(
-            "Unable to find `functiontrace-server` in the current $PATH.",
-            file=sys.stderr,
-        )
-        print(
-            "See https://functiontrace.com#installation for installation instructions.",
-            file=sys.stderr,
-        )
-        sys.exit(1)
-
     # Generate a temp directory to store our wrappers in.  We'll temporarily
     # add this directory to our path.
     tempdir = tempfile.mkdtemp(prefix="py-functiontrace")
@@ -79,6 +65,9 @@ def main():
         parser.print_help()
         sys.exit(1)
 
+    # Load our config and verify the license before we trace.
+    rust.load_config()
+
     # Ignore ourselves, keeping sys.argv looking reasonable as the child script
     # will expect it to be sane.
     sys.argv[:] = args.script
